#ifndef EXPR_H
#define EXPR_H

#include <iostream>
#include <string>

class Expr {
public:
  virtual std::string toString() const {
    return "<expr>";
  }

  friend std::ostream& operator<<(std::ostream& out, const Expr& expr) {
    return out << expr.toString();
  }

  virtual int eval() = 0;

  virtual Expr* copy() const { return nullptr; }

  virtual ~Expr() {}
};

// override
// why `public`?
struct Number : public Expr {
  int value;

  Number(int value) : value(value) {}

  virtual int eval() override {
    return value;
  }

  std::string toString() const override {
    return std::to_string(value);
  }

  Expr* copy() const override { return new Number(*this); }
};

struct BinaryExpr : public Expr {
protected:
  Expr * lhs;
  char op;
  Expr * rhs;
public:

  // TODO: Add deep copy: need a method. Why?
  
  BinaryExpr(Expr * lhs, char op, Expr * rhs) : lhs(lhs->copy()), op(op), rhs(rhs->copy()) {}

  std::string toString() const override {
    return "(" + lhs->toString() + op + rhs->toString() + ")";
  }

  BinaryExpr(const BinaryExpr& that) : lhs(that.lhs->copy()),
                                       op(that.op),
                                       rhs(that.rhs->copy())
  {
  }

  BinaryExpr& operator=(const BinaryExpr& that) = delete;

  ~BinaryExpr() {
    delete lhs;
    delete rhs;
  }
};

struct Add : public BinaryExpr {
  Add(Expr * lhs, Expr * rhs) : BinaryExpr(lhs, '+', rhs) {}

  virtual int eval() override {
    return this->lhs->eval() + this->rhs->eval();
  }

  Expr* copy() const override { return new Add(lhs->copy(),
                                               rhs->copy()); }
};

struct Mul : public BinaryExpr {
  Mul(Expr * lhs, Expr * rhs) : BinaryExpr(lhs, '*', rhs) {}

  virtual int eval() override {
    return lhs->eval() * rhs->eval();
  }

  Expr* copy() const override { return new Mul(lhs->copy(),
                                               rhs->copy()); }
};

#endif  // EXPR_H
